from .bases import *

__copyright__ = "Copyright (C) 2020 Yuichiro Smith"
__version__ = "0.2.3"
__license__ = "Apache License 2.0"
__author__ = "Yuichiro Smith"
__author_email__ = "contact@yu-smith.com"
__url__ = "http://github.com/aeroworks-io/python-ddd"



from abc import ABC, abstractmethod

from .abstract import JsonSerializable, Restoreable, Generatable


class PrimitiveBase(JsonSerializable, Restoreable, Generatable, ABC):
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    @abstractmethod
    def validate(cls, v):
        ...


class Primitive(PrimitiveBase):
    coerce = True

    @classmethod
    def validate(cls, v):
        if isinstance(v, cls):
            return v
        if cls.coerce:
            # noinspection PyArgumentList
            return cls(v)
        raise TypeError(f"expected {repr(v)} to be a instance of {cls}")

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"

    def __json__(self):
        return self

    @classmethod
    def __restore__(cls, value):
        # noinspection PyArgumentList
        return cls(value)

    @classmethod
    def __generate__(cls):
        return cls()



from ulid import ULID


from ..common import BaseClass, PrimitiveBase


class ID(PrimitiveBase, ULID):
    coerce = True

    def __repr__(self):
        return f"{self.__class__.__name__}('{str(self)}')"

    def __json__(self):
        return str(self)

    @classmethod
    def validate(cls, v):
        if isinstance(v, cls):
            return v
        if cls.coerce:
            return cls.__restore__(v)
        raise TypeError(f"expected {repr(v)} to be a instance of {cls}")

    @classmethod
    def __generate__(cls):
        return cls()

    @classmethod
    def __restore__(cls, value):
        return cls.from_str(str(value))


class Entity(BaseClass):
    id: ID

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.id == other.id


class Value(BaseClass):
    class Config:
        allow_mutation = False



from abc import ABC, abstractmethod
from typing import TypeVar, Type, Any, Optional

T = TypeVar("T")
GenericAlias = type(Optional[int])


class Restoreable(ABC):
    @classmethod
    @abstractmethod
    def __restore__(cls: Type[T], value) -> T:
        ...


class Generatable(ABC):
    @classmethod
    @abstractmethod
    def __generate__(cls: Type[T]) -> T:
        ...


class JsonSerializable(ABC):
    @abstractmethod
    def __json__(self: T) -> Any:
        ...


def find_abstract_constraint(annotation, abstract: Type[T]) -> Optional[T]:
    if type(annotation) is GenericAlias:
        for ann in annotation.__args__:
            constraint = find_abstract_constraint(ann, abstract)
            if constraint:
                return constraint
    try:
        if issubclass(annotation, abstract):
            return annotation
    except TypeError:
        return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dddpy', 'dddpy.bases', 'dddpy.bases.common', 'dddpy.bases.domain']

package_data = \
{'': ['*']}

install_requires = \
['Faker>=4.0.0,<7.0.0',
 'Inject>=4.3.1,<5.0.0',
 'pydantic>=1.7.2,<2.0.0',
 'python-ulid>=1.0.2,<2.0.0']

setup_kwargs = {
    'name': 'dddpy',
    'version': '0.2.3',
    'description': 'A framework to support ddd python projects',
    'long_description': '# dddpy\n[![CI](https://github.com/aeroworks-io/dddpy/workflows/CI/badge.svg)](https://github.com/aeroworks-io/dddpy/actions?query=workflow%3ACI)  \n[![codecov](https://codecov.io/gh/aeroworks-io/dddpy/branch/main/graph/badge.svg?token=BOO43Q8GIF)](https://codecov.io/gh/aeroworks-io/dddpy)\n[![PyPI version](https://badge.fury.io/py/dddpy.svg)](https://badge.fury.io/py/dddpy)\n\nA Framework to support ddd python projects\n',
    'author': 'Yuichiro Smith',
    'author_email': 'contact@yu-smith.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aeroworks-io/dddpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Optional

from pydantic import BaseModel, BaseConfig, ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import FieldInfo as _FieldInfo, Undefined
from pydantic.typing import NoArgAnyCallable
from pydantic.utils import ROOT_KEY

from .abstract import (
    JsonSerializable,
    Restoreable,
    Generatable,
    find_abstract_constraint,
)
from .primitive import PrimitiveBase, Primitive


class BaseClass(Restoreable, Generatable, JsonSerializable, BaseModel):
    class Config(BaseConfig):
        json_encoders = {JsonSerializable: lambda ins: ins.__json__()}

    def __init__(self, *args, **kwargs):
        kw = tuple(self.__class__.__annotations__)
        if len(args) > len(kw):
            raise TypeError(
                f"{self.__class__.__name__}.__init__() takes {len(kw)} positional argument but {len(args)} were "
                f"given"
            )
        super().__init__(**dict(zip(kw, args)), **kwargs)

    @classmethod
    def parse_obj(cls, obj: Any):
        obj = cls._enforce_dict_if_root(obj)
        if not isinstance(obj, dict):
            try:
                obj = dict(obj)
            except (TypeError, ValueError) as e:
                exc = TypeError(
                    f"{cls.__name__} expected dict not {obj.__class__.__name__}"
                )
                raise ValidationError([ErrorWrapper(exc, loc=ROOT_KEY)], cls) from e
        return super().parse_obj(
            {
                k: typ.__restore__(v)
                if (
                    typ := find_abstract_constraint(cls.__annotations__[k], Restoreable)
                )
                else v
                for k, v in obj.items()
            }
        )

    def __json__(self) -> str:
        return self.json()

    @classmethod
    def __restore__(cls, value: Any):
        return cls.parse_obj(value)

    @classmethod
    def __generate__(cls):
        # noinspection PyArgumentList
        return cls(
            **{
                key: typ.__generate__()
                for key in cls.__annotations__
                if (
                    typ := find_abstract_constraint(
                        cls.__annotations__[key], Generatable
                    )
                )
            }
        )


class FieldInfo(_FieldInfo):
    ...


# noinspection PyPep8Naming
def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional[NoArgAnyCallable] = None,
    alias: str = None,
    title: str = None,
    description: str = None,
    const: bool = None,
    gt: float = None,
    ge: float = None,
    lt: float = None,
    le: float = None,
    multiple_of: float = None,
    min_items: int = None,
    max_items: int = None,
    min_length: int = None,
    max_length: int = None,
    regex: str = None,
    **extra: Any,
) -> Any:
    if default is not Undefined and default_factory is not None:
        raise ValueError("cannot specify both default and default_factory")

    return FieldInfo(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        min_items=min_items,
        max_items=max_items,
        min_length=min_length,
        max_length=max_length,
        regex=regex,
        **extra,
    )
